#include "stdlib.h"
#include "RetroEngine.h"

CRetroEngine::CRetroEngine():w(0), h(0), screen(nullptr), hblank(nullptr) {
	int x;
	this->SetLayer(0, false, 0, 0);
	this->SetLayer(1, false, 0, 0);
	this->SetLayer(2, false, 0, 0);
	this->SetLayer(3, false, 0, 0);
	for (x = 0; x < RE_NUM_SPRITES; x++) {
		this->ClearSprite(x);
	}
}

int CRetroEngine::SetResolution(int inWidth, int inHeight) {
	if (inWidth <= 0) return 0;
	if (inHeight <= 0) return 0;

	if (this->screen != nullptr) free(this->screen);

	this->screen = (uint8 *)malloc(4*inWidth*inHeight*sizeof(uint8));
	this->w = inWidth;
	this->h = inHeight;

	return 1;
}

int CRetroEngine::Draw() {
	int x,y,l,i,j;							// loop iterators
	int xoffs, yoffs;						// x/y scroll offset
	int tileNo;								// index of tile
	uint8 cx, cy;							// character x, y (needed for flipping)
	uint8 pixInd;							// pixel index for sampling a character
	uint32* pal;								// pointer into palette
	uint32* tscr;							// screen pointer
	re_sprite sp[RE_NUM_SPRITES];			// local sprite cache
	re_sprite_bounds spb[RE_NUM_SPRITES];	// sprite bounds
	int nSprites = 0;						// number of active sprites
	re_sprite ssp[RE_NUM_SPRITES];			// scanline sprite cache
	re_sprite_bounds sspb[RE_NUM_SPRITES];	// scanilne sprite bounds
	int nScanlineSprites = 0;				// number of sprites on this scanline

	if (this->screen == nullptr) return 0;

	this->lockout = true;	// set the lockout flag so that the HBLANK callback can't alter private vars

	// copy sprite table to local memory
	// (this is why we lock-out sprite editing during hblank - sprites are sorted at the start
	// of the frame and are not re-sorted each scanline)
	nSprites=0;
	for (x = 0; x < RE_NUM_SPRITES; x++) if (this->sprites[x].data & RE_SPRITE_ENABLE_MASK) {
		sp[nSprites++] = this->sprites[x];
	}

	// sort sprites
	this->SortSprites(sp, nSprites);

	// calculate sprite bounds - speeds up intersection tests later
	for (x = 0; x < nSprites; x++) {
		spb[x].left = (sp[x].exdata & 0x01FF) - 32;
		spb[x].top = (sp[x].data & 0x01FF) - 32;
		spb[x].right = spb[x].left + ((((sp[x].data >> 14) & 0x03) + 1) << 3) - 1;
		spb[x].bottom = spb[x].top + ((((sp[x].data >> 12) & 0x03) + 1) << 3) - 1;
	}

	// go to start of screen
	tscr = (uint32 *)this->screen;

	// iterate through all scanlines
	for (y=0; y<this->h; y++) {
		// call HBLANK callback
		if (this->hblank != nullptr) {
			this->hblank(this, y, this->hblankPayload);
		}

		// build list of sprites on this scanline
		nScanlineSprites = 0;
		for (i = 0; i < nSprites; i++) {
			if (!(spb[i].top > y || spb[i].bottom < y)) {
				ssp[nScanlineSprites] = sp[i];
				sspb[nScanlineSprites] = spb[i];
				nScanlineSprites++;
			}
		}

		// if blending is enabled on the top layer
		if (this->layer[0]&RE_LAYER_ENABLE_FLAG && ((this->layer[0]>>28) & 0x07) != RE_BLEND_OFF) {

			// iterate through all columns
			for (x=0; x<this->w; x++) {

				// storage for layer0 RGBA color
				uint8 l0_red = 0;
				uint8 l0_green = 0;
				uint8 l0_blue = 0;
				uint8 l0_alpha = 0;

				// reset to top background layer
				int StartLayer = 0;
				int StartSprite = 0;
				uint8 bg_col		= 0;				// translucent
				uint8 bg_priority	= 1;				// low priority
				uint8 bg_layer		= RE_NUM_LAYERS -1;	// bottom layer (if nothing is found, fill with palette[0])

				// LAYER 0 BACKGROUND COLOR
				l = 0;
				
				// get x/y coords on tilemap
				xoffs = (((this->layer[l]>>0 ) & 0x1FF) + x) & (RE_TILEMAP_SIZE_X*8-1);
				yoffs = (((this->layer[l]>>16) & 0x1FF) + y) & (RE_TILEMAP_SIZE_Y*8-1);
				
				// find the tile
				tileNo = (((yoffs & ~0x07) << 3) | (xoffs >> 3));
				
				// get tile coords
				cx = xoffs & 0x07;
				cy = yoffs & 0x07;
				if (this->tilemap[l][tileNo] & RE_TILE_FLIP_X) cx = 7-cx;
				if (this->tilemap[l][tileNo] & RE_TILE_FLIP_Y) cy = 7-cy;
				
				// get the color
				bg_col = this->SampleChar(this->tilemap[l][tileNo] & 0x03FF, cx, cy);

				// if not translucent
				if (bg_col != 0) {
				
					// record color
					bg_col |= (this->tilemap[l][tileNo] >> 6) & 0x70;
					bg_priority = this->tilemap[l][tileNo] >> 15;
					bg_layer = 0;
				}

				// LAYER 0 SPRITE COLOR
				uint8 sprite_col		= 0;				// translucent
				uint8 sprite_priority	= 1;				// low priority
				uint8 sprite_layer		= RE_NUM_LAYERS -1; // bottom layer

				// iterate through sprites on this scanline (but only until we hit layer 1)
				for (i=0;i<nScanlineSprites && ((ssp[i].exdata>>10)&0x0003)==0;i++) {
					uint16 tileno;
					uint8 tcol = 0;		// default to translucent
					uint8 sx;
					uint8 sy;

					// if this sprite intersects the pixel
					if (!(x<sspb[i].left || x>sspb[i].right)) {
						sx = (ssp[i].data & RE_SPRITE_FLIP_X)?(sspb[i].right - x):(x - sspb[i].left);
						sy = (ssp[i].data & RE_SPRITE_FLIP_Y)?(sspb[i].bottom - y):(y - sspb[i].top);
						tcol = this->SampleChar((ssp[i].data >> 16)&0x03FF, sx, sy);
					}
					// if this sprite pixel is NOT translucent
					if (tcol != 0) {
						sprite_col = tcol | ((ssp[i].exdata >> 8) & 0xf0);							// move to [palette][color]
						sprite_priority = (ssp[i].exdata & RE_SPRITE_PRIORITY_MASK)?1:0;	// record sprite priority
						sprite_layer = 0;													// record layer
						i = nScanlineSprites;												// move to end of list
					}
				}

				// choose sprite or background color
				if (sprite_col == 0) {							// no sprite found
					pal = this->palette + bg_col;
				} else if (bg_col == 0) {						// no background found
					pal = this->palette + sprite_col;
				} else if (bg_layer == sprite_layer) {			// sprte and background on same layer ...
					if (bg_priority < sprite_priority) {			// ... priority to background
						pal = this->palette + bg_col;
					} else {										// ... priority to sprite
						pal = this->palette + sprite_col;
					}
				} else if (bg_layer < sprite_layer) {			// background on lower layer than sprite
					pal = this->palette + bg_col;
				} else {										// sprite on lower layer than background
					pal = this->palette + sprite_col;
				}

				
				// if not translucent
				if (bg_col != 0 || sprite_col != 0) {
					// record color
					l0_red = (*pal)>>16;
					l0_green = (*pal)>>8;
					l0_blue = (*pal);
					l0_alpha = ((this->layer[0]>>9)&0x7F);
				} else {
					l0_alpha = 0;
				}

				StartLayer = 1;
				StartSprite = 0;

				// iterate through remaining layers
				for (l=StartLayer;l<RE_NUM_LAYERS;l++) if (this->layer[l]&RE_LAYER_ENABLE_FLAG) {
				
					// get x/y coords on tilemap
					xoffs = (((this->layer[l]>>0 ) & 0x1FF) + x) & (RE_TILEMAP_SIZE_X*8-1);
					yoffs = (((this->layer[l]>>16) & 0x1FF) + y) & (RE_TILEMAP_SIZE_Y*8-1);
				
					// find the tile
					tileNo = (((yoffs & ~0x07) << 3) | (xoffs >> 3));
				
					// get tile coords
					cx = xoffs & 0x07;
					cy = yoffs & 0x07;
					if (this->tilemap[l][tileNo] & RE_TILE_FLIP_X) cx = 7-cx;
					if (this->tilemap[l][tileNo] & RE_TILE_FLIP_Y) cy = 7-cy;
				
					// get the color
					bg_col = this->SampleChar(this->tilemap[l][tileNo] & 0x03FF, cx, cy);
				
					// if not translucent
					if (bg_col != 0) {
				
						// record color
						bg_col |= (this->tilemap[l][tileNo] >> 6) & 0x70;
						bg_priority = this->tilemap[l][tileNo] >> 15;
						bg_layer = l;
					
						// exit loop
						l = RE_NUM_LAYERS;
					}
				}

				// get sprite color

				sprite_col = 0;
				sprite_priority = 1;
				sprite_layer = RE_NUM_LAYERS -1;

				// find the first sprite NOT on layer0 (remember, these are pre-sorted)
				if (nScanlineSprites > 0) {
					while (StartSprite < nScanlineSprites && ((ssp[StartSprite].exdata>>10)&0x0003)==0) {
						StartSprite++;
					}
				}

				// iterate through remaining sprites
				for (i = StartSprite; i < nScanlineSprites; i++) {
					uint16 tileno;
					uint8 tcol = 0;
					uint8 sx;
					uint8 sy;

					// if this sprite intersects this pixel, get color
					if (!(x<sspb[i].left || x>sspb[i].right)) {
						sx = (ssp[i].data & RE_SPRITE_FLIP_X)?(sspb[i].right - x):(x - sspb[i].left);
						sy = (ssp[i].data & RE_SPRITE_FLIP_Y)?(sspb[i].bottom - y):(y - sspb[i].top);
						tcol = this->SampleChar((ssp[i].data >> 16)&0x03FF, sx, sy);
					}
					// if this pixel is NOT translucent
					if (tcol != 0) {
						sprite_col = tcol | ((ssp[i].exdata >> 8) & 0xf0);							// get [palette][color]
						sprite_priority = (ssp[i].exdata & RE_SPRITE_PRIORITY_MASK)?1:0;	// get sprite priority
						sprite_layer = (ssp[i].exdata >> 10) & 0x0003;						// get sprite layer
						i = nScanlineSprites;												// jump to end of list
					}
				}

				// choose sprite or background color
				if (sprite_col == 0) {							// no sprite found
					pal = this->palette + bg_col;
				} else if (bg_col == 0) {						// no background found
					pal = this->palette + sprite_col;
				} else if (bg_layer == sprite_layer) {			// sprte and background on same layer ...
					if (bg_priority < sprite_priority) {			// ... priority to background
						pal = this->palette + bg_col;
					} else {										// ... priority to sprite
						pal = this->palette + sprite_col;
					}
				} else if (bg_layer < sprite_layer) {			// background on lower layer than sprite
					pal = this->palette + bg_col;
				} else {										// sprite on lower layer than background
					pal = this->palette + sprite_col;
				}

				// copy color
				uint8 tr = (*pal)>>16;
				uint8 tg = (*pal)>>8;
				uint8 tb = (*pal);

				// blend in the layer0 color
				uint8 blend_mode = (this->layer[0] >> 28) & 0x07;
				uint8 l0_ialpha = 127-l0_alpha;
				if (blend_mode == RE_BLEND_BASIC) {
					tr = ((l0_red*l0_alpha) + (tr*l0_ialpha))>>7;
					tg = ((l0_green*l0_alpha) + (tg*l0_ialpha))>>7;
					tb = ((l0_blue*l0_alpha) + (tb*l0_ialpha))>>7;
				} else if (blend_mode == RE_BLEND_ADD) {
					uint16 tmp;
					tmp = ((l0_red*l0_alpha) >> 7) + tr;
					tr = tmp>255?255:tmp;
					tmp = ((l0_green*l0_alpha) >> 7) + tg;
					tg = tmp>255?255:tmp;
					tmp = ((l0_blue*l0_alpha) >> 7) + tb;
					tb = tmp>255?255:tmp;
				} else if (blend_mode == RE_BLEND_MULT) {
					tr = ((l0_red*l0_alpha + 255*l0_ialpha)*tr)>>15;
					tg = ((l0_green*l0_alpha + 255*l0_ialpha)*tg)>>15;
					tb = ((l0_blue*l0_alpha + 255*l0_ialpha)*tb)>>15;
				}

				*tscr = tb | (tg<<8) | (tr<<16);

				// move to next pixel
				tscr ++;
			}

		} else {
			// iterate through all columns
			for (x=0; x<this->w; x++) {
				// reset to top background layer
				int StartLayer = 0;
				uint8 bg_col		= 0;				// translucent
				uint8 bg_priority	= 1;				// low priority
				uint8 bg_layer		= RE_NUM_LAYERS -1;	// bottom layer (if nothing is found, fill with palette[0])

				// iterate through remaining layers
				for (l=StartLayer;l<RE_NUM_LAYERS;l++) if (this->layer[l]&RE_LAYER_ENABLE_FLAG) {
				
					// get x/y coords on tilemap
					xoffs = (((this->layer[l]>>0 ) & 0x1FF) + x) & (RE_TILEMAP_SIZE_X*8-1);
					yoffs = (((this->layer[l]>>16) & 0x1FF) + y) & (RE_TILEMAP_SIZE_Y*8-1);
				
					// find the tile
					tileNo = (((yoffs & ~0x07) << 3) | (xoffs >> 3));
				
					// get tile coords
					cx = xoffs & 0x07;
					cy = yoffs & 0x07;
					if (this->tilemap[l][tileNo] & RE_TILE_FLIP_X) cx = 7-cx;
					if (this->tilemap[l][tileNo] & RE_TILE_FLIP_Y) cy = 7-cy;
				
					// get the color
					bg_col = this->SampleChar(this->tilemap[l][tileNo] & 0x03FF, cx, cy);
				
					// if not translucent
					if (bg_col != 0) {
				
						// record color
						bg_col |= (this->tilemap[l][tileNo] >> 6) & 0x70;
						bg_priority = this->tilemap[l][tileNo] >> 15;
						bg_layer = l;
					
						// exit loop
						l = RE_NUM_LAYERS;
					}
				}

				// get sprite color
				uint8 sprite_col		= 0;				// translucent
				uint8 sprite_priority	= 1;				// low priority
				uint8 sprite_layer		= RE_NUM_LAYERS -1; // bottom layer

				// iterate through all sprites
				for (i = 0; i < nScanlineSprites; i++) {
					uint16 tileno;
					uint8 tcol = 0;
					uint8 sx;
					uint8 sy;

					// if this sprite intersects this pixel, get sprite color
					if (!(x<sspb[i].left || x>sspb[i].right)) {
						sx = (ssp[i].data & RE_SPRITE_FLIP_X)?(sspb[i].right - x):(x - sspb[i].left);
						sy = (ssp[i].data & RE_SPRITE_FLIP_Y)?(sspb[i].bottom - y):(y - sspb[i].top);
						tcol = this->SampleChar((ssp[i].data >> 16)&0x03FF, sx, sy);
					}
					// if this pixel is NOT translucent
					if (tcol != 0) {
						sprite_col = tcol | ((ssp[i].exdata >> 8) & 0xf0);					// store [palette][color]
						sprite_priority = (ssp[i].exdata & RE_SPRITE_PRIORITY_MASK)?1:0;	// store sprite priority
						sprite_layer = (ssp[i].exdata >> 10) & 0x0003;						// store sprite layer
						i = nScanlineSprites;												// move to end
					}
				}

				// choose sprite or background color
				if (sprite_col == 0) {							// no sprite found
					pal = this->palette + bg_col;
				} else if (bg_col == 0) {						// no background found
					pal = this->palette + sprite_col;
				} else if (bg_layer == sprite_layer) {			// sprte and background on same layer ...
					if (bg_priority < sprite_priority) {			// ... priority to background
						pal = this->palette + bg_col;
					} else {										// ... priority to sprite
						pal = this->palette + sprite_col;
					}
				} else if (bg_layer < sprite_layer) {			// background on lower layer than sprite
					pal = this->palette + bg_col;
				} else {										// sprite on lower layer than background
					pal = this->palette + sprite_col;
				}

				// copy color
				*tscr = *pal;

				// move to next pixel
				tscr ++;
			}
		}
	}


	this->lockout = false;	// unset the lockout flag, giving the main procedure the ability to change private vars

	return 1;
}

int CRetroEngine::SetLayer (int inLayerNumber, bool inEnable, unsigned int inXPos, unsigned int inYPos, uint8 inBlendMode, uint8 inAlpha) {
	// ensure valid layer number
	if (inLayerNumber < 0 || inLayerNumber >= 4) return 0;

	// create a temporary variable and clear everything
	uint32 tmp = 0x00000000;

	// set enable flag
	if (inEnable) tmp |= RE_LAYER_ENABLE_FLAG;

	// set blend mode & alpha
	tmp |= (inBlendMode & 0x07) << 28;
	tmp |= (inAlpha & 0x7F) << 9;

	// set scroll offset
	tmp |= (inXPos & 0x1FF) << 0;
	tmp |= (inYPos & 0x1FF) << 16;

	// save data to layer memory
	this->layer[inLayerNumber] = tmp;

	return 1;
}

// This feature is not yet implemented. Instead, create a solid color tile and set the entire
// tilemap to that tile.
int CRetroEngine::SetLayerSolid (int inLayerNumber, bool inEnable, uint8 inRed, uint8 inGreen, uint8 inBlue, uint8 inBlendMode, uint8 inAlpha) {
	// solid colors are only viable on the top layer
	if (inLayerNumber != 0) return 0;
	return 0;
}


int CRetroEngine::SetSprite(int inSpriteIndex, bool enable, uint16 tileNo, uint8 w, uint8 h, uint8 layer, uint8 priority, uint16 xpos, uint16 ypos, bool flipx, bool flipy, uint8 pal) {
	if (inSpriteIndex < 0 || inSpriteIndex >= RE_NUM_SPRITES) return 0;

	// we store xpos and ypos as unsigned integers offset by the maximum sprite size so that they can go off-screen in any direction
	xpos += 32;
	ypos += 32;
	if (xpos<0) xpos = 0;
	if (ypos<0) ypos = 0;
	if (xpos>this->w+32) xpos = this->w+32;
	if (ypos>this->h+32) ypos = this->h+32;
	
	uint32 tmp_d = 0x00000000;		// tempoarary sprite data (32 bits)
	uint16 tmp_e = 0x0000;			// tempoarary sprite exdata (16 bits)

	// exdata - [palette 4b][layer 2b][priority 1b][xpos 9b]
	tmp_e = (xpos & 0x01FF) | ((priority & 0x01)<<9) | ((layer & 0x03) << 10) | ((pal & 0x0F)<<12);
	// data - [blank 7b][enable 1b][tile 9b][width 2b][height 2b][blank 1b][flipx 1b][flipy 1b][ypos 9b]
	tmp_d = (ypos & 0x01FF) | (flipx?RE_SPRITE_FLIP_X:0) | (flipy?RE_SPRITE_FLIP_Y:0) | ((h&0x03)<<12) | ((w&0x03)<<14) | ((tileNo&0x01FF)<<16) | (enable?RE_SPRITE_ENABLE_MASK:0);

	this->sprites[inSpriteIndex].data = tmp_d;
	this->sprites[inSpriteIndex].exdata = tmp_e;

	return 1;
}

void CRetroEngine::SortSprites(re_sprite *buff, int count) {
	// let's just do a quick and dirty selection sort for now
	re_sprite lowest;
	int x, y, ind;
	for (x = 0; x < count - 1; x++) {
		ind = x;
		lowest = buff[x];
		for (y = x; y < count; y++) {
			// re_sprite.exdata stores layer, priority, and xpos in its lower 12 bits
			// so we can sort by layer, then by priority, then left-to-right by
			// masking out palette a doing a simple integer comparison.
			if ((buff[y].exdata & 0x0FFF) < (lowest.exdata & 0x0FFF)) {
				ind = y;
				lowest = buff[y];
			}
		}
		if (ind != x) {
			buff[ind] = buff[x];
			buff[x] = lowest;
		}
	}
}