// for Visual C++
#pragma once
// for everybody else ...
#ifndef _RETROENGINE_H
#define _RETROENGINE_H

// RetroEngine Types
#ifndef uint8
typedef unsigned char uint8;
#endif
#ifndef uint16
typedef unsigned short uint16;
#endif
#ifndef uint32
typedef unsigned int uint32;
#endif

typedef uint16 re_tile;
typedef struct {
	uint32 data;
	uint16 exdata;
} re_sprite;
typedef uint32 re_layer;
typedef struct {
	int left;
	int top;
	int right;
	int bottom;
} re_sprite_bounds;

// RetroEngine Constants
//   These constants should not be changed unless you're prepared
//   to rewrite the draw routine. For example, because there are 4
//   layers, we have 2 bits allocated to store layer numbers.
//   Changing this constant to 8 won't allocate an additional bit
//   and layers 4-7 would be inaccessible.
const int RE_NUM_LAYERS =				4;
const int RE_TILEMAP_SIZE_X =			64;
const int RE_TILEMAP_SIZE_Y =			64;
const int RE_NUM_SPRITES =				256;

const uint32 RE_LAYER_ENABLE_FLAG =		1 << 31;

const uint8 RE_BLEND_OFF =				0;
const uint8 RE_BLEND_BASIC =			1;
const uint8 RE_BLEND_ADD =				2;
const uint8 RE_BLEND_MULT =				3;

const uint16 RE_TILE_FLIP_X =			1 << 14;
const uint16 RE_TILE_FLIP_Y =			1 << 13;
const uint16 RE_TILE_PRIORITY_MASK =	1 << 15;

const uint32 RE_SPRITE_FLIP_X =			1 << 10;
const uint32 RE_SPRITE_FLIP_Y =			1 << 9;
const uint16 RE_SPRITE_PRIORITY_MASK =	1 << 9;
const uint32 RE_SPRITE_ENABLE_MASK =	1 << 26;

#define re_make_tile(index, palette, flipx, flipy, priority) ((index)&0x03ff) | (((palette)&0x0007)<<10) | (((priority)&0x0001)<<15) | (flipx?RE_TILE_FLIP_X:0) | (flipy?RE_TILE_FLIP_Y:0)

class CRetroEngine {
	// DATA
private:
	int			w, h;						// screen width and height
	uint8		*screen;					// raw screen data
	bool		lockout;					// internal lockout flag for hblanking
	re_layer	layer[RE_NUM_LAYERS];		// layer data
	re_sprite	sprites[RE_NUM_SPRITES];	// sprite data
	
	void (*hblank)(CRetroEngine *, int, void *);	// hblank callback
	void *hblankPayload;							// pointer to data structure that's used by hblank
													// routine to get data from the caller

public:
	re_tile		tilemap[RE_NUM_LAYERS][RE_TILEMAP_SIZE_X*RE_TILEMAP_SIZE_Y];
	uint32		palette[16*16];			// ARGB for 16 palettes of 15 colors each (some memory waste here)
	uint8		chr_memory[8*8*1024/2];		// 1024 8x8 sprites at 1/2 byte per pixel

	// FUNCTIONS

private:
	void SortSprites (re_sprite *buff, int count);

	// gets the color from chr_memory for a given index, x offset, y offset
	inline uint8 SampleChar(const uint16 &charIndex, const uint8 &x, const uint8 &y) {
		uint8 *offset_ptr = this->chr_memory + ((charIndex & 0x0F) << 2) + ((charIndex >> 4) << 9);
		uint8 col = offset_ptr[(y<<6) + (x>>1)];		// get the byte of memory containing color
		col = col >> (((~x) & 0x01)<<2);				// if even pixel, take high nybble
		col = col & 0x0F;								// make sure we only return 1 nybble
		return col;
	};
		
public:
	CRetroEngine();
	int SetResolution(int inWidth, int inHeight);
	int Draw();
	int SetLayer (int inLayerNumber, bool inEnable, unsigned int inXPos, unsigned int inYPos, uint8 inBlendMode=0x00, uint8 inAlpha=0x00);
	int SetLayerSolid (int inLayerNumber, bool inEnable, uint8 inRed, uint8 inGreen, uint8 inBlue, uint8 inBlendMode=0x00, uint8 inAlpha=0x00);
	int SetSprite (int inSpriteIndex, bool enable, uint16 tileNo, uint8 w, uint8 h, uint8 layer, uint8 priority, uint16 xpos, uint16 ypos, bool flipx, bool flipy, uint8 pal);
	
	// returns screen pointer so that it can be blitted onto window
	inline uint8 *GetScreen() {
		return this->screen;
	};

	// sets a sprite to default values and disable
	inline int ClearSprite (int inSpriteIndex) {
		return this->SetSprite(inSpriteIndex, false, 0, 0, 0, 0, 0, 0, 0, false, false, 0);
	};

	// sets the hblank callback function
	inline void SetHBlank(void(*inHB)(CRetroEngine *, int, void *), void *payload = nullptr) {
		if (inHB != nullptr) this->hblank = inHB;
		this->hblankPayload = payload;
	};

	// clears the hblank callback function
	inline void ClearHBlank() {
		this->hblank = nullptr;
	};
};



#endif // _RETROENGINE_H