/*

	Typing Academy
	by: Matt Pierce
	November 2018

	This is a basic typing tutor game for children.
	I created this because I inteded to give my
	daughter a Raspberry Pi for Hanukkah but 
	couldn't find any typing tutor games (except
	tux typing) aimed at young kids.

*/
#include <iostream>
#include <math.h>
#include <string.h>
#include "GlobalDataStruct.h"
#include "SDL2/SDL.h"
#include "sdl_virtual_screen.h"
#include "RetroEngine.h"
#include "TitleScreen.h"
#include "MainMenu.h"
#include "GameplayLoop.h"
#include "level1.h"

using namespace std;

// GLOBAL CONSTANTS - resolution settings
#ifdef FULLSCREEN_DEF
const bool FullScreen = true;
#else 
const bool FullScreen = false;
#endif
const int InternalScreenWidth = 320;
const int InternalScreenHeight = 180;
const int WindowedScale = 3;
const int DefaultScreenWidth = InternalScreenWidth * WindowedScale;
const int DefaultScreenHeight = InternalScreenHeight * WindowedScale;

// GLOBAL FUNCTIONS - forward declarations
void InitGame(GlobalData *globals);
void FinishGame(GlobalData *globals);
void LoadChrMemory(uint8* chr_memory);
void DecompressSNES(uint8 *chr_memory);
void LoadPal(uint32 *pal);


int main(int argc, char *argv[]) {
	bool GameRunning = true;
	SDL_Event evt;

	GlobalData			globals;		// global values - lives here, pointer gets passed around
	CTitleScreen		titleScreen;	// title screen game state
	CMainMenu			menuLoop;		// main menu game state

	// Initialize SDL and global stuff
	InitGame(&globals);

	// set the virtual screen resolution
	globals.screen.Initialize(InternalScreenWidth, InternalScreenHeight, globals.win);

	// set the RetroEngine resolution
	globals.re.SetResolution(InternalScreenWidth, InternalScreenHeight);

	// set pacemaker to give us a 60 frame per second pulse
	globals.pacemaker.interval = SDL_GetPerformanceFrequency()/60;

	// load tiles and decompress SNES bitplanes
	LoadChrMemory(globals.re.chr_memory);
	DecompressSNES(globals.re.chr_memory);

	// load palette
	LoadPal(globals.re.palette);

	// title screen state
	titleScreen.Init(&globals);
	titleScreen.Loop();

	// main menu state
	menuLoop.Init(&globals);
	menuLoop.Loop();

	// shut down SDL
	FinishGame(&globals);
	return 0;
}


/**
<summary>Initializes SDL window according to global constants.</summary>
*/
void InitGame(GlobalData *globals) {

	// start SDL
	SDL_Init(SDL_INIT_EVERYTHING);

	if (FullScreen) {
		// When creating a fullscreen window, don't change the current desktop mode. Just make a big window.
		globals->win = SDL_CreateWindow("RetroEngine Test", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 0, 0, SDL_WINDOW_FULLSCREEN_DESKTOP);
		SDL_ShowCursor(SDL_FALSE);	// hide cursor when fullscreen
	} else {
		globals->win = SDL_CreateWindow("RetroEngine Test", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, DefaultScreenWidth, DefaultScreenHeight, SDL_WINDOW_SHOWN);
	}
	
	// set the title here so that we don't have to worry about the titles above matching
	SDL_SetWindowTitle(globals->win, "Typing Academy");
}

/**
<summary>Shuts down SDL and exits</summary>
*/
void FinishGame(GlobalData *globals) {
	SDL_DestroyWindow(globals->win);
	SDL_Quit();
	exit(0); // should I really do this here? Should I return and exit?
}

/**
<summary>Loads binary sprites into chr memory</summary>
<param name="chr_memory">a pointer to RetroEngine chr_memory buffer</param>
*/
void LoadChrMemory(uint8* chr_memory) {
	FILE *fp;
	size_t chunkSize = 8*8*1024/2;
	size_t bytesRead;

	// we have a relative path, but let's get an absolute path
	// using a cross-platform method
	char *data_path = SDL_GetBasePath();
	char filename[2048] = "";
	if (strstr(data_path, "\\Release\\")!=NULL || strstr(data_path, "\\Debug\\")!=NULL) {
		// We're running from inside the Visual Studio IDE, so do nothing -
		// let the Visual Studio project resolve relative paths.
	} else {
		// we're not running from inside Visual Studio, so figure out where we are
		strcat(filename, data_path);
	}
	strcat(filename, "assets/typing.chr");

	// open file
	fp = fopen(filename, "rb");

	if (fp) {
		// read it all in one go
		bytesRead = fread(chr_memory, sizeof(uint8), chunkSize, fp);
		fclose(fp);
	}

}
/**
 * Loads the palette and converts to ARGB for 32-bit systems
 */
void LoadPal(uint32 *pal) {
	FILE *fp;
	size_t bytesRead;
	uint8 *inputBuffer, *tib;
	int i;

	// convert relative path to absolute path
	char *data_path = SDL_GetBasePath();
	char filename[2048] = "";
	if (strstr(data_path, "\\Release\\")!=NULL || strstr(data_path, "\\Debug\\")!=NULL) {
		// We're running from inside the Visual Studio IDE, so do nothing -
		// let the Visual Studio project resolve relative paths.
	} else {
		// we're not running from inside Visual Studio, so figure out where we are
		strcat(filename, data_path);
	}
	strcat(filename, "assets/typing.pal");

	// open file
	fp = fopen(filename, "rb");

	if (fp) {
		// create input buffer and read in the palette as 768 bytes
		inputBuffer = (uint8 *)malloc(768);
		bytesRead = fread(inputBuffer, sizeof(uint8), 768, fp);

		// get a pointer to the front of the buffer
		tib = inputBuffer;

		// iterate through all 256 colors
		for (i=0;i<256;i++) {

			// convert 3 bytes into a single 32-bit unsigned integer
			*pal = tib[0]<<16 | tib[1]<<8 | tib[2];

			// advance to next color
			pal++;
			tib += 3;
		}

		// clean up
		free(inputBuffer);
		fclose(fp);
	}
}

/**
 * Converts SNES bit-planes into simple discrete values
 */
void DecompressSNES(uint8 *chr_memory) {
	uint8 *tmp_memory;
	uint8 *readptr, *writeptr;
	uint32 b0, b1, b2, b3;
	int i, j;

	// allocate temporary storage
	tmp_memory = (uint8*)malloc(32768);
	for (i = 0; i < 1024; i++) {
		// go to the sprite index in each map
		readptr = chr_memory + 32*i;
		writeptr = tmp_memory + ((i&0x0f)<<2) + ((i>>4)<<9);
		for (j = 0; j < 8; j++) {
			// get bitplane 0 & 1
			b0 = readptr[0];
			b1 = readptr[1];
			// de-leave planes 0 and 1
			b0 = (b0 | (b0<<4)) & 0x0f0f;
			b0 = (b0 | (b0<<2)) & 0x3333;
			b0 = (b0 | (b0<<1)) & 0x5555;
			b1 = (b1 | (b1<<4)) & 0x0f0f;
			b1 = (b1 | (b1<<2)) & 0x3333;
			b1 = (b1 | (b1<<1)) & 0x5555;
			b0 = b0 | (b1<<1);
			// get bitplane 2 & 3
			b2 = readptr[16];
			b3 = readptr[17];
			// de-leave planes 2 and 3
			b2 = (b2 | (b2<<4)) & 0x0f0f;
			b2 = (b2 | (b2<<2)) & 0x3333;
			b2 = (b2 | (b2<<1)) & 0x5555;
			b3 = (b3 | (b3<<4)) & 0x0f0f;
			b3 = (b3 | (b3<<2)) & 0x3333;
			b3 = (b3 | (b3<<1)) & 0x5555;
			b2 = b2 | (b3<<1);
			// combine as: 22002200220022002200220022002200
			b0 = (b0 | (b0<<8)) & 0x00ff00ff00ff00ff;
			b0 = (b0 | (b0<<4)) & 0x0f0f0f0f0f0f0f0f;
			b0 = (b0 | (b0<<2)) & 0x3333333333333333;
			b2 = (b2 | (b2<<8)) & 0x00ff00ff00ff00ff;
			b2 = (b2 | (b2<<4)) & 0x0f0f0f0f0f0f0f0f;
			b2 = (b2 | (b2<<2)) & 0x3333333333333333;
			b0 = b0 | (b2<<2);
			// save data (this is clumsy but we can't guarantee endian-ness)
			writeptr[0] = (b0>>24) & 0xff;
			writeptr[1] = (b0>>16) & 0xff;
			writeptr[2] = (b0>>8)  & 0xff;
			writeptr[3] = (b0>>0)  & 0xff;
			// advance to next
			readptr += 2;
			writeptr += 64;
		}
	}

	// copy memory
	memcpy(chr_memory, tmp_memory, 32768);

	// free temp memory
	free(tmp_memory);
}

